"""System Bridge: GUI"""
import asyncio
import subprocess
import sys
from threading import Event, Thread
from typing import Callable, Optional

import async_timeout
from systembridgeshared.base import Base
from systembridgeshared.exceptions import ConnectionErrorException
from systembridgeshared.settings import Settings
from systembridgeshared.websocket_client import WebSocketClient


class StoppableThread(Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the thread."""
        super().__init__(*args, **kwargs)
        self._stop_event = Event()

    def stop(self) -> None:
        """Stop the thread."""
        self._stop_event.set()

    def stopped(self) -> bool:
        """Return if the thread is stopped."""
        return self._stop_event.is_set()


class GUI(Base):
    """GUI"""

    def __init__(
        self,
        settings: Settings,
    ):
        """Initialize"""
        super().__init__()
        self._settings = settings

        self._name = "GUI"
        self._process: Optional[subprocess.Popen] = None
        self._stopping = False
        self._thread: Optional[StoppableThread] = None

    async def _start(  # pylint: disable=keyword-arg-before-vararg
        self,
        failed_callback: Optional[Callable[[], None]],
        attempt: int = 1,
        command: str = "main",
        *args,
    ) -> None:
        """Start the GUI"""
        if attempt > 2:
            self._logger.error("Failed to start GUI after 2 attempts")
            if failed_callback is not None:
                failed_callback()
            return
        if command == "main":
            self._logger.info(
                "Test WebSocket connection before starting GUI. Attempt #%s", attempt
            )
            websocket_client = WebSocketClient(self._settings)
            try:
                async with async_timeout.timeout(20):
                    await websocket_client.connect()
                    await websocket_client.close()
            except ConnectionErrorException:
                self._logger.warning(
                    "Could not connect to WebSocket. Retrying in 5 seconds"
                )
                await asyncio.sleep(5)
                await self._start(
                    failed_callback,
                    attempt + 1,
                    command,
                    *args,
                )
                return
            except asyncio.TimeoutError:
                self._logger.warning(
                    "Connection timeout to WebSocket. Retrying in 5 seconds"
                )
                await asyncio.sleep(5)
                await self._start(
                    failed_callback,
                    attempt + 1,
                    command,
                    *args,
                )
                return

        self._logger.info("Connection test passed")

        pgm_args = [
            sys.executable,
            "-m",
            "systembridgegui",
            command,
            *args,
        ]

        self._name = command

        self._logger.info("Starting GUI: %s", pgm_args)
        with subprocess.Popen(pgm_args) as self._process:
            self._logger.info("GUI started with PID: %s", self._process.pid)
            if (exit_code := self._process.wait()) != 0:
                if not self._stopping:
                    self._logger.error("GUI exited with code: %s", exit_code)
                    await self._start(
                        failed_callback,
                        attempt + 1,
                        command,
                        *args,
                    )
                    return
            self._logger.info("GUI exited with code: %s", exit_code)

    def _start_gui_sync(  # pylint: disable=keyword-arg-before-vararg
        self,
        failed_callback: Optional[Callable[[], None]],
        command: str = "main",
        *args,
    ) -> None:
        """Start the GUI in a synchronous thread"""
        asyncio.run(
            self._start(
                failed_callback,
                1,
                command,
                *args,
            )
        )

    async def start(  # pylint: disable=keyword-arg-before-vararg
        self,
        failed_callback: Optional[Callable[[], None]],
        command: str = "main",
        *args,
    ) -> None:
        """Start the GUI"""
        self._thread = StoppableThread(
            target=self._start_gui_sync,
            args=(
                failed_callback,
                command,
                *args,
            ),
        )
        self._thread.start()
        self._stopping = False

    def stop(self) -> None:
        """Stop the GUI"""
        self._logger.info("Stopping GUI: %s", self._name)
        self._stopping = True
        if self._process is not None:
            self._process.terminate()
            self._process.wait()
            self._process = None
            self._logger.info("GUI stopped")
        if self._thread is not None:
            self._thread.stop()



"""System Bridge: GPU"""
from __future__ import annotations

from typing import Optional

from systembridgeshared.base import Base
from systembridgeshared.common import make_key
from systembridgeshared.database import Database
from systembridgeshared.models.database_data_sensors import (
    Sensors as SensorsDatabaseModel,
)


class GPU(Base):
    """GPU"""

    def get_gpus(
        self,
        database: Database,
    ) -> list[str]:
        """Get GPUs"""
        gpus = []
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "gpu" in item.hardware_type.lower()
                and item.hardware_name not in gpus
            ):
                gpus.append(item.hardware_name)
        return gpus

    def core_clock(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU core clock"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "gpu" in item.hardware_type.lower()
                and "clock" in item.type.lower()
                and "core" in item.name.lower()
                and make_key(item.hardware_name) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU core clock: %s = %s",
                    item.key,
                    item.value,
                )
                return item.value
        return None

    def core_load(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU core load"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "gpu" in item.hardware_type.lower()
                and "load" in item.type.lower()
                and "core" in item.name.lower()
                and make_key(item.hardware_name) == gpu_key
            ):
                self._logger.debug("Found GPU core load: %s = %s", item.key, item.value)
                return item.value
        return None

    def fan_speed(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU fan speed"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "gpu" in item.hardware_type.lower()
                and "fan" in item.type.lower()
                and make_key(item.hardware_name) == gpu_key
            ):
                self._logger.debug("Found GPU fan speed: %s = %s", item.key, item.value)
                return item.value
        return None

    def memory_clock(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU memory clock"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "gpu" in item.hardware_type.lower()
                and "clock" in item.type.lower()
                and "memory" in item.name.lower()
                and make_key(item.hardware_name) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU memory clock: %s = %s",
                    item.key,
                    item.value,
                )
                return item.value
        return None

    def memory_load(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU memory load"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "gpu" in item.hardware_type.lower()
                and "load" in item.type.lower()
                and "memory" in item.name.lower()
                and make_key(item.hardware_name) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU memory load: %s = %s",
                    item.key,
                    item.value,
                )
                return item.value
        return None

    def memory_free(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU memory free"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "gpu" in item.hardware_type.lower()
                and "memory" in item.name.lower()
                and "free" in item.name.lower()
                and make_key(item.hardware_name) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU memory free: %s = %s",
                    item.key,
                    item.value,
                )
                return item.value
        return None

    def memory_used(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU memory used"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "gpu" in item.hardware_type.lower()
                and "memory" in item.name.lower()
                and "used" in item.name.lower()
                and make_key(item.hardware_name) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU memory used: %s = %s",
                    item.key,
                    item.value,
                )
                return item.value
        return None

    def memory_total(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU memory total"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "gpu" in item.hardware_type.lower()
                and "memory" in item.name.lower()
                and "total" in item.name.lower()
                and make_key(item.hardware_name) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU memory total: %s = %s",
                    item.key,
                    item.value,
                )
                return item.value
        return None

    def power(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU power usage"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "gpu" in item.hardware_type.lower()
                and "power" in item.type.lower()
                and make_key(item.hardware_name) == gpu_key
            ):
                self._logger.debug("Found GPU power: %s = %s", item.key, item.value)
                return item.value
        return None

    def temperature(
        self,
        database: Database,
        gpu_key: str,
    ) -> Optional[float]:
        """GPU temperature"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "gpu" in item.hardware_type.lower()
                and "temperature" in item.type.lower()
                and "core" in item.name.lower()
                and make_key(item.hardware_name) == gpu_key
            ):
                self._logger.debug(
                    "Found GPU temperature: %s = %s",
                    item.key,
                    item.value,
                )
                return item.value
        return None



"""System Bridge: Modules Update"""
import asyncio
from collections.abc import Awaitable, Callable

from systembridgeshared.base import Base
from systembridgeshared.database import Database

from .battery.update import BatteryUpdate
from .cpu.update import CPUUpdate
from .disk.update import DiskUpdate
from .display.update import DisplayUpdate
from .gpu.update import GPUUpdate
from .memory.update import MemoryUpdate
from .network.update import NetworkUpdate
from .sensors.update import SensorsUpdate
from .system.update import SystemUpdate


class Update(Base):
    """Modules Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__()
        self._database = database  # pylint: disable=duplicate-code

        self._classes = [
            {"name": "battery", "cls": BatteryUpdate(self._database)},
            {"name": "disk", "cls": DiskUpdate(self._database)},
            {"name": "system", "cls": SystemUpdate(self._database)},
        ]
        self._classes_frequent = [
            {"name": "cpu", "cls": CPUUpdate(self._database)},
            {"name": "display", "cls": DisplayUpdate(self._database)},
            {"name": "gpu", "cls": GPUUpdate(self._database)},
            {"name": "memory", "cls": MemoryUpdate(self._database)},
            {"name": "network", "cls": NetworkUpdate(self._database)},
        ]

    async def _update(
        self,
        class_obj: dict,
        updated_callback: Callable[[str], Awaitable[None]],
    ) -> None:
        """Update"""
        await class_obj["cls"].update_all_data()
        await updated_callback(class_obj["name"])

    async def update_data(
        self,
        updated_callback: Callable[[str], Awaitable[None]],
    ) -> None:
        """Update Data"""
        self._logger.info("Update data")

        tasks = [self._update(cls, updated_callback) for cls in self._classes]
        await asyncio.gather(*tasks)

        self._logger.info("Finished updating data")

    async def update_frequent_data(
        self,
        updated_callback: Callable[[str], Awaitable[None]],
    ) -> None:
        """Update Data"""
        self._logger.info("Update frequent data")

        sensors_update = SensorsUpdate(self._database)
        await sensors_update.update_all_data()
        await updated_callback("sensors")

        tasks = [self._update(cls, updated_callback) for cls in self._classes_frequent]
        await asyncio.gather(*tasks)

        self._logger.info("Finished updating frequent data")



"""System Bridge: Update Disk"""
import asyncio
from json import dumps

from systembridgeshared.database import Database
from systembridgeshared.models.database_data import Disk as DatabaseModel

from . import Disk
from ..base import ModuleUpdateBase


class DiskUpdate(ModuleUpdateBase):
    """Disk Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database)
        self._disk = Disk()

    async def update_io_counters(self) -> None:
        """Update IO counters"""
        for key, value in self._disk.io_counters()._asdict().items():
            self._database.update_data(
                DatabaseModel,
                DatabaseModel(
                    key=f"io_counters_{key}",
                    value=value,
                ),
            )

    async def update_io_counters_per_disk(self) -> None:
        """Update IO counters per disk"""
        for key, value in self._disk.io_counters_per_disk().items():  # type: ignore
            for subkey, subvalue in value._asdict().items():
                self._database.update_data(
                    DatabaseModel,
                    DatabaseModel(
                        key=f"io_counters_per_disk_{key}_{subkey}",
                        value=subvalue,
                    ),
                )

    async def update_partitions(self) -> None:
        """Update partitions"""
        device_list = []
        partition_list = []
        for partition in self._disk.partitions():
            if partition.device not in device_list:
                device_list.append(partition.device)
            partition_list.append(partition.mountpoint)
            for key, value in partition._asdict().items():
                self._database.update_data(
                    DatabaseModel,
                    DatabaseModel(
                        key=f"partitions_{partition.mountpoint}_{key}",
                        value=value,
                    ),
                )
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="devices",
                value=dumps(device_list),
            ),
        )
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="partitions",
                value=dumps(partition_list),
            ),
        )

    async def update_usage(self) -> None:
        """Update usage"""
        for partition in self._disk.partitions():
            if data := self._disk.usage(partition.mountpoint):
                for key, value in data._asdict().items():
                    self._database.update_data(
                        DatabaseModel,
                        DatabaseModel(
                            key=f"usage_{partition.mountpoint}_{key}",
                            value=value,
                        ),
                    )

    async def update_all_data(self) -> None:
        """Update data"""
        await asyncio.gather(
            *[
                self.update_io_counters(),
                self.update_io_counters_per_disk(),
                self.update_partitions(),
                self.update_usage(),
            ]
        )



"""System Bridge: Server Handler - Remote Bridge"""
from systembridgeshared.database import Database
from systembridgeshared.models.database_data_remote_bridge import RemoteBridge


def get_remote_bridges(
    database: Database,
) -> list[RemoteBridge]:
    """Get all remote bridges."""
    return database.get_data(RemoteBridge)



"""System Bridge: WebSocket handler"""
import os
from collections.abc import Callable
from json import JSONDecodeError
from uuid import uuid4

from fastapi import WebSocket
from starlette.websockets import WebSocketDisconnect
from systembridgeshared.base import Base
from systembridgeshared.const import (
    EVENT_BASE,
    EVENT_DATA,
    EVENT_DIRECTORIES,
    EVENT_EVENT,
    EVENT_FILE,
    EVENT_FILES,
    EVENT_ID,
    EVENT_MESSAGE,
    EVENT_MODULE,
    EVENT_MODULES,
    EVENT_PATH,
    EVENT_SETTING,
    EVENT_SUBTYPE,
    EVENT_TYPE,
    EVENT_URL,
    EVENT_VALUE,
    EVENT_VERSIONS,
    SETTING_AUTOSTART,
    SUBTYPE_BAD_API_KEY,
    SUBTYPE_BAD_DIRECTORY,
    SUBTYPE_BAD_FILE,
    SUBTYPE_BAD_JSON,
    SUBTYPE_BAD_PATH,
    SUBTYPE_BAD_REQUEST,
    SUBTYPE_LISTENER_ALREADY_REGISTERED,
    SUBTYPE_LISTENER_NOT_REGISTERED,
    SUBTYPE_MISSING_KEY,
    SUBTYPE_MISSING_MODULES,
    SUBTYPE_MISSING_PATH_URL,
    SUBTYPE_MISSING_TEXT,
    SUBTYPE_MISSING_TITLE,
    SUBTYPE_UNKNOWN_EVENT,
    TYPE_APPLICATION_UPDATE,
    TYPE_APPLICATION_UPDATING,
    TYPE_DATA_GET,
    TYPE_DATA_LISTENER_REGISTERED,
    TYPE_DATA_LISTENER_UNREGISTERED,
    TYPE_DATA_UPDATE,
    TYPE_DIRECTORIES,
    TYPE_ERROR,
    TYPE_EXIT_APPLICATION,
    TYPE_FILE,
    TYPE_FILES,
    TYPE_GET_DATA,
    TYPE_GET_DIRECTORIES,
    TYPE_GET_FILE,
    TYPE_GET_FILES,
    TYPE_GET_REMOTE_BRIDGES,
    TYPE_GET_REMOTE_BRIDGES_RESULT,
    TYPE_GET_SETTING,
    TYPE_GET_SETTINGS,
    TYPE_KEYBOARD_KEY_PRESSED,
    TYPE_KEYBOARD_KEYPRESS,
    TYPE_KEYBOARD_TEXT,
    TYPE_KEYBOARD_TEXT_SENT,
    TYPE_NOTIFICATION,
    TYPE_NOTIFICATION_SENT,
    TYPE_OPEN,
    TYPE_OPENED,
    TYPE_POWER_HIBERNATE,
    TYPE_POWER_HIBERNATING,
    TYPE_POWER_LOCK,
    TYPE_POWER_LOCKING,
    TYPE_POWER_LOGGINGOUT,
    TYPE_POWER_LOGOUT,
    TYPE_POWER_RESTART,
    TYPE_POWER_RESTARTING,
    TYPE_POWER_SHUTDOWN,
    TYPE_POWER_SHUTTINGDOWN,
    TYPE_POWER_SLEEP,
    TYPE_POWER_SLEEPING,
    TYPE_REGISTER_DATA_LISTENER,
    TYPE_SETTING_RESULT,
    TYPE_SETTING_UPDATED,
    TYPE_SETTINGS_RESULT,
    TYPE_UNREGISTER_DATA_LISTENER,
    TYPE_UPDATE_REMOTE_BRIDGE,
    TYPE_UPDATE_REMOTE_BRIDGE_RESULT,
    TYPE_UPDATE_SETTING,
)
from systembridgeshared.database import TABLE_MAP, Database
from systembridgeshared.models.data import DataDict
from systembridgeshared.models.database_data_remote_bridge import RemoteBridge
from systembridgeshared.models.get_data import GetData
from systembridgeshared.models.get_setting import GetSetting
from systembridgeshared.models.keyboard_key import KeyboardKey
from systembridgeshared.models.keyboard_text import KeyboardText
from systembridgeshared.models.media_get_file import MediaGetFile
from systembridgeshared.models.media_get_files import MediaGetFiles
from systembridgeshared.models.notification import Notification
from systembridgeshared.models.open_path import OpenPath
from systembridgeshared.models.open_url import OpenUrl
from systembridgeshared.models.register_data_listener import RegisterDataListener
from systembridgeshared.models.request import Request
from systembridgeshared.models.response import Response
from systembridgeshared.models.update import Update as UpdateModel
from systembridgeshared.models.update_setting import UpdateSetting
from systembridgeshared.settings import SECRET_API_KEY, Settings
from systembridgeshared.update import Update

from ..autostart import autostart_disable, autostart_enable
from ..modules.listeners import Listeners
from ..server.keyboard import keyboard_keypress, keyboard_text
from ..server.media import get_directories, get_file, get_files
from ..server.open import open_path, open_url
from ..server.power import hibernate, lock, logout, restart, shutdown, sleep
from ..server.remote_bridge import get_remote_bridges


class WebSocketHandler(Base):
    """WebSocket handler"""

    def __init__(
        self,
        database: Database,
        settings: Settings,
        listeners: Listeners,
        implemented_modules: list[str],  # pylint: disable=unsubscriptable-object
        websocket: WebSocket,
        callback_exit_application: Callable[[], None],
        callback_open_gui: Callable[[str, str], None],
    ) -> None:
        """Initialize"""
        super().__init__()
        self._database = database
        self._settings = settings
        self._listeners = listeners
        self._implemented_modules = implemented_modules
        self._websocket = websocket
        self._callback_exit_application = callback_exit_application
        self._callback_open_gui = callback_open_gui
        self._active = True

    async def _send_response(
        self,
        response: Response,
    ) -> None:
        """Send response"""
        if not self._active:
            return
        message = response.dict()
        self._logger.debug("Sending message: %s", message)
        await self._websocket.send_json(message)

    async def _data_changed(
        self,
        module: str,
        data: DataDict,
    ) -> None:
        """Data changed"""
        if module not in self._implemented_modules:
            self._logger.info("Data module %s not in registered modules", module)
            return
        await self._send_response(
            Response(
                **{
                    EVENT_TYPE: TYPE_DATA_UPDATE,
                    EVENT_MESSAGE: "Data changed",
                    EVENT_MODULE: module,
                    EVENT_DATA: data,
                }
            )
        )

    async def _handler(
        self,
        listener_id: str,
    ) -> None:
        """Handler"""
        # Loop until the connection is closed
        while self._active:
            try:
                data = await self._websocket.receive_json()
                request = Request(**data)
            except JSONDecodeError as error:
                message = f"Invalid JSON: {error}"
                self._logger.error(message)
                await self._send_response(
                    Response(
                        **{
                            EVENT_TYPE: TYPE_ERROR,
                            EVENT_SUBTYPE: SUBTYPE_BAD_JSON,
                            EVENT_MESSAGE: message,
                        }
                    )
                )
                continue
            except ValueError as error:
                message = f"Invalid request: {error}"
                self._logger.error(message)
                await self._send_response(
                    Response(
                        **{
                            EVENT_TYPE: TYPE_ERROR,
                            EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                            EVENT_MESSAGE: message,
                        }
                    )
                )
                continue

            self._logger.info("Received: %s", request.event)

            if request.api_key != self._settings.get_secret(SECRET_API_KEY):
                self._logger.warning("Invalid api-key")
                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_ERROR,
                            EVENT_SUBTYPE: SUBTYPE_BAD_API_KEY,
                            EVENT_MESSAGE: "Invalid api-key",
                        }
                    )
                )
                continue
            if request.event == TYPE_APPLICATION_UPDATE:
                try:
                    model = UpdateModel(**data)
                except ValueError as error:
                    message = f"Invalid request: {error}"
                    self._logger.warning(message)
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                                EVENT_MESSAGE: message,
                            }
                        )
                    )
                    continue
                versions = Update().update(
                    model.version,
                    wait=False,
                )
                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_APPLICATION_UPDATING,
                            EVENT_MESSAGE: "Updating application",
                            EVENT_VERSIONS: versions,
                        }
                    )
                )
            elif request.event == TYPE_EXIT_APPLICATION:
                self._callback_exit_application()
                self._logger.info("Exit application called")
            elif request.event == TYPE_KEYBOARD_KEYPRESS:
                try:
                    model = KeyboardKey(**data)
                except ValueError as error:
                    message = f"Invalid request: {error}"
                    self._logger.warning(message)
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                                EVENT_MESSAGE: message,
                            }
                        )
                    )
                    continue
                if model.key is None:
                    self._logger.warning("No key provided")
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_KEY,
                                EVENT_MESSAGE: "No key provided",
                            }
                        )
                    )
                    continue

                try:
                    keyboard_keypress(model.key)
                except ValueError as err:
                    self._logger.warning(err.args[0])
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_KEY,
                                EVENT_MESSAGE: "Invalid key",
                            }
                        )
                    )
                    continue

                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_KEYBOARD_KEY_PRESSED,
                            EVENT_MESSAGE: "Key pressed",
                            EVENT_ID: listener_id,
                            "key": model.key,
                        }
                    )
                )
            elif request.event == TYPE_KEYBOARD_TEXT:
                try:
                    model = KeyboardText(**data)
                except ValueError as error:
                    message = f"Invalid request: {error}"
                    self._logger.warning(message)
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                                EVENT_MESSAGE: message,
                            }
                        )
                    )
                    continue
                if model.text is None:
                    self._logger.warning("No text provided")
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_TEXT,
                                EVENT_MESSAGE: "No text provided",
                            }
                        )
                    )
                    continue

                keyboard_text(model.text)

                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_KEYBOARD_TEXT_SENT,
                            EVENT_MESSAGE: "Key pressed",
                            EVENT_ID: listener_id,
                            "text": model.text,
                        }
                    )
                )
            elif request.event == TYPE_NOTIFICATION:
                try:
                    model = Notification(**data)
                except ValueError as error:
                    message = f"Invalid request: {error}"
                    self._logger.warning(message)
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                                EVENT_MESSAGE: message,
                            }
                        )
                    )
                    continue
                if model.title is None:
                    self._logger.warning("No title provided")
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_TITLE,
                                EVENT_MESSAGE: "No title provided",
                            }
                        )
                    )
                    continue

                self._callback_open_gui("notification", model.json())

                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_NOTIFICATION_SENT,
                            EVENT_MESSAGE: "Notification sent",
                        }
                    )
                )
            elif request.event == TYPE_OPEN:
                if "path" in data:
                    try:
                        model = OpenPath(**data)
                    except ValueError as error:
                        message = f"Invalid request: {error}"
                        self._logger.warning(message)
                        await self._send_response(
                            Response(
                                **{
                                    EVENT_ID: request.id,
                                    EVENT_TYPE: TYPE_ERROR,
                                    EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                                    EVENT_MESSAGE: message,
                                }
                            )
                        )
                        continue
                    open_path(model.path)
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_OPENED,
                                EVENT_MESSAGE: "Path opened",
                                EVENT_PATH: model.path,
                            }
                        )
                    )
                    continue
                if "url" in data:
                    try:
                        model = OpenUrl(**data)
                    except ValueError as error:
                        message = f"Invalid request: {error}"
                        self._logger.warning(message)
                        await self._send_response(
                            Response(
                                **{
                                    EVENT_ID: request.id,
                                    EVENT_TYPE: TYPE_ERROR,
                                    EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                                    EVENT_MESSAGE: message,
                                }
                            )
                        )
                        continue
                    open_url(model.url)
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_OPENED,
                                EVENT_MESSAGE: "URL opened",
                                EVENT_URL: model.url,
                            }
                        )
                    )
                    continue

                self._logger.warning("No path or url provided")
                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_ERROR,
                            EVENT_SUBTYPE: SUBTYPE_MISSING_PATH_URL,
                            EVENT_MESSAGE: "No path or url provided",
                        }
                    )
                )
            elif request.event == TYPE_REGISTER_DATA_LISTENER:
                try:
                    model = RegisterDataListener(**data)
                except ValueError as error:
                    message = f"Invalid request: {error}"
                    self._logger.warning(message)
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                                EVENT_MESSAGE: message,
                            }
                        )
                    )
                    continue
                if model.modules is None or len(model.modules) == 0:
                    self._logger.warning("No modules provided")
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_MODULES,
                                EVENT_MESSAGE: "No modules provided",
                            }
                        )
                    )
                    continue

                self._logger.info(
                    "Registering data listener: %s - %s",
                    listener_id,
                    model.modules,
                )

                if await self._listeners.add_listener(
                    listener_id,
                    self._data_changed,
                    model.modules,
                ):
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_LISTENER_ALREADY_REGISTERED,
                                EVENT_MESSAGE: "Listener already registered with this connection",
                                EVENT_MODULES: model.modules,
                            }
                        )
                    )
                    continue

                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_DATA_LISTENER_REGISTERED,
                            EVENT_MESSAGE: "Data listener registered",
                            EVENT_MODULES: model.modules,
                        }
                    )
                )
            elif request.event == TYPE_UNREGISTER_DATA_LISTENER:
                self._logger.info("Unregistering data listener %s", listener_id)

                if not self._listeners.remove_listener(listener_id):
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_LISTENER_NOT_REGISTERED,
                                EVENT_MESSAGE: "Listener not registered with this connection",
                            }
                        )
                    )
                    continue

                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_DATA_LISTENER_UNREGISTERED,
                            EVENT_MESSAGE: "Data listener unregistered",
                        }
                    )
                )
            elif request.event == TYPE_GET_DATA:
                try:
                    model = GetData(**data)
                except ValueError as error:
                    message = f"Invalid request: {error}"
                    self._logger.warning(message)
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                                EVENT_MESSAGE: message,
                            }
                        )
                    )
                    continue
                if model.modules is None or len(model.modules) == 0:
                    self._logger.warning("No modules provided")
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_MISSING_MODULES,
                                EVENT_MESSAGE: "No modules provided",
                            }
                        )
                    )
                    continue
                self._logger.info("Getting data: %s", model.modules)

                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_DATA_GET,
                            EVENT_MESSAGE: "Getting data",
                            EVENT_MODULES: model.modules,
                        }
                    )
                )

                for module in model.modules:
                    table = TABLE_MAP.get(module)
                    data = self._database.get_data_dict(table)
                    if data is not None:
                        await self._send_response(
                            Response(
                                **{
                                    EVENT_ID: request.id,
                                    EVENT_TYPE: TYPE_DATA_UPDATE,
                                    EVENT_MESSAGE: "Data received",
                                    EVENT_MODULE: module,
                                    EVENT_DATA: data,
                                }
                            )
                        )
            elif request.event == TYPE_GET_DIRECTORIES:
                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_DIRECTORIES,
                            EVENT_DIRECTORIES: get_directories(self._settings),
                        }
                    )
                )
            elif request.event == TYPE_GET_FILES:
                try:
                    model = MediaGetFiles(**data)
                except ValueError as error:
                    message = f"Invalid request: {error}"
                    self._logger.warning(message)
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                                EVENT_MESSAGE: message,
                            }
                        )
                    )
                    continue

                root_path = None
                for item in get_directories(self._settings):
                    if item["key"] == model.base:
                        root_path = item["path"]
                        break

                if root_path is None or not os.path.exists(root_path):
                    self._logger.warning("Cannot find base path")
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_PATH,
                                EVENT_MESSAGE: "Cannot find base path",
                                EVENT_BASE: model.base,
                            }
                        )
                    )
                    continue

                path = (
                    os.path.join(root_path, model.path)
                    if model.path is not None
                    else root_path
                )

                self._logger.info(
                    "Getting files: %s - %s - %s",
                    model.base,
                    model.path,
                    path,
                )

                if not os.path.exists(path):
                    self._logger.warning("Cannot find path")
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_PATH,
                                EVENT_MESSAGE: "Cannot find path",
                                EVENT_PATH: path,
                            }
                        )
                    )
                    continue
                if not os.path.isdir(path):
                    self._logger.warning("Path is not a directory")
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_DIRECTORY,
                                EVENT_MESSAGE: "Path is not a directory",
                                EVENT_PATH: path,
                            }
                        )
                    )
                    continue

                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_FILES,
                            EVENT_FILES: get_files(self._settings, model.base, path),
                            EVENT_PATH: path,
                        }
                    )
                )
            elif request.event == TYPE_GET_FILE:
                try:
                    model = MediaGetFile(**data)
                except ValueError as error:
                    message = f"Invalid request: {error}"
                    self._logger.warning(message)
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                                EVENT_MESSAGE: message,
                            }
                        )
                    )
                    continue

                root_path = None
                for item in get_directories(self._settings):
                    if item["key"] == model.base:
                        root_path = item["path"]
                        break

                if root_path is None or not os.path.exists(root_path):
                    self._logger.warning("Cannot find base path")
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_PATH,
                                EVENT_MESSAGE: "Cannot find base path",
                                EVENT_BASE: model.base,
                            }
                        )
                    )
                    continue

                path = os.path.join(root_path, model.path)

                self._logger.info(
                    "Getting file: %s - %s - %s",
                    model.base,
                    model.path,
                    path,
                )

                if not os.path.exists(path):
                    self._logger.warning("Cannot find path")
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_PATH,
                                EVENT_MESSAGE: "Cannot find path",
                                EVENT_PATH: path,
                            }
                        )
                    )
                    continue
                if not os.path.isfile(path):
                    self._logger.warning("Path is not a file")
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_FILE,
                                EVENT_MESSAGE: "Path is not a file",
                                EVENT_PATH: path,
                            }
                        )
                    )
                    continue

                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_FILE,
                            EVENT_FILE: get_file(root_path, path),
                            EVENT_PATH: path,
                        }
                    )
                )
            elif request.event == TYPE_GET_SETTINGS:
                self._logger.info("Getting settings")
                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_SETTINGS_RESULT,
                            EVENT_MESSAGE: "Got settings",
                            EVENT_DATA: self._settings.get_all(),
                        }
                    )
                )
            elif request.event == TYPE_GET_SETTING:
                try:
                    model = GetSetting(**data)
                except ValueError as error:
                    message = f"Invalid request: {error}"
                    self._logger.warning(message)
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                                EVENT_MESSAGE: message,
                            }
                        )
                    )
                    continue

                self._logger.info("Getting setting: %s", model.setting)

                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_SETTING_RESULT,
                            EVENT_MESSAGE: "Got setting",
                            EVENT_SETTING: model.setting,
                            EVENT_DATA: self._settings.get(model.setting),
                        }
                    )
                )
            elif request.event == TYPE_GET_REMOTE_BRIDGES:
                self._logger.info("Getting remote bridges")
                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_GET_REMOTE_BRIDGES_RESULT,
                            EVENT_MESSAGE: "Got remote bridges",
                            EVENT_DATA: get_remote_bridges(self._database),
                        }
                    )
                )
            elif request.event == TYPE_UPDATE_REMOTE_BRIDGE:
                try:
                    model = RemoteBridge(**data)
                except ValueError as error:
                    message = f"Invalid request: {error}"
                    self._logger.warning(message)
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                                EVENT_MESSAGE: message,
                            }
                        )
                    )
                    continue

                self._logger.info("Remote bridge: %s", model)

                model = self._database.update_remote_bridge(model)

                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_UPDATE_REMOTE_BRIDGE_RESULT,
                            EVENT_MESSAGE: "Data updated",
                            EVENT_DATA: model,
                        }
                    )
                )
            elif request.event == TYPE_UPDATE_SETTING:
                try:
                    model = UpdateSetting(**data)
                except ValueError as error:
                    message = f"Invalid request: {error}"
                    self._logger.warning(message)
                    await self._send_response(
                        Response(
                            **{
                                EVENT_ID: request.id,
                                EVENT_TYPE: TYPE_ERROR,
                                EVENT_SUBTYPE: SUBTYPE_BAD_REQUEST,
                                EVENT_MESSAGE: message,
                            }
                        )
                    )
                    continue

                self._logger.info(
                    "Setting setting %s to: %s",
                    model.setting,
                    model.value,
                )

                self._settings.set(model.setting, model.value)

                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_SETTING_UPDATED,
                            EVENT_MESSAGE: "Setting updated",
                            EVENT_SETTING: model.setting,
                            EVENT_VALUE: model.value,
                        }
                    )
                )

                if model.setting != SETTING_AUTOSTART:
                    continue
                self._logger.info("Setting autostart to %s", model.value)
                if model.value is True:
                    autostart_enable()
                else:
                    autostart_disable()
            elif request.event == TYPE_POWER_SLEEP:
                self._logger.info("Sleeping")
                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_POWER_SLEEPING,
                            EVENT_MESSAGE: "Sleeping",
                        }
                    )
                )
                sleep()
            elif request.event == TYPE_POWER_HIBERNATE:
                self._logger.info("Sleeping")
                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_POWER_HIBERNATING,
                            EVENT_MESSAGE: "Hiibernating",
                        }
                    )
                )
                hibernate()
            elif request.event == TYPE_POWER_RESTART:
                self._logger.info("Sleeping")
                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_POWER_RESTARTING,
                            EVENT_MESSAGE: "Restarting",
                        }
                    )
                )
                restart()
            elif request.event == TYPE_POWER_SHUTDOWN:
                self._logger.info("Sleeping")
                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_POWER_SHUTTINGDOWN,
                            EVENT_MESSAGE: "Shutting down",
                        }
                    )
                )
                shutdown()
            elif request.event == TYPE_POWER_LOCK:
                self._logger.info("Locking")
                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_POWER_LOCKING,
                            EVENT_MESSAGE: "Locking",
                        }
                    )
                )
                lock()
            elif request.event == TYPE_POWER_LOGOUT:
                self._logger.info("Logging out")
                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_POWER_LOGGINGOUT,
                            EVENT_MESSAGE: "Logging out",
                        }
                    )
                )
                logout()
            else:
                self._logger.warning("Unknown event: %s", request.event)
                await self._send_response(
                    Response(
                        **{
                            EVENT_ID: request.id,
                            EVENT_TYPE: TYPE_ERROR,
                            EVENT_SUBTYPE: SUBTYPE_UNKNOWN_EVENT,
                            EVENT_MESSAGE: "Unknown event",
                            EVENT_EVENT: request.event,
                        }
                    )
                )

    async def handler(self) -> None:
        """Handler"""
        listener_id = str(uuid4())
        try:
            await self._handler(listener_id)
        except (ConnectionError, WebSocketDisconnect) as error:
            self._logger.info("Connection closed: %s", error)
        finally:
            self._logger.info("Unregistering data listener %s", listener_id)
            self._listeners.remove_listener(listener_id)

    def set_active(
        self,
        active: bool,
    ) -> None:
        """Set active"""
        self._active = active



"""System Bridge: Update Display"""
import asyncio
from json import dumps
from typing import Optional

from systembridgeshared.common import make_key
from systembridgeshared.database import Database
from systembridgeshared.models.database_data import Display as DatabaseModel

from ..base import ModuleUpdateBase
from . import Display


class DisplayUpdate(ModuleUpdateBase):
    """Display Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database)
        self._display = Display()

    async def update_name(
        self,
        display_key: str,
        display_name: str,
    ) -> None:
        """Update name"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{display_key}_name",
                value=display_name,
            ),
        )

    async def update_pixel_clock(
        self,
        display_key: str,
        value: Optional[float] = None,
    ) -> None:
        """Update pixel clock"""
        if value is None:
            value = self._display.sensors_pixel_clock(self._database, display_key)
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{display_key}_pixel_clock",
                value=str(value) if value is not None else None,
            ),
        )

    async def update_refresh_rate(
        self,
        display_key: str,
        value: Optional[float] = None,
    ) -> None:
        """Update refresh rate"""
        if value is None:
            value = self._display.sensors_refresh_rate(self._database, display_key)
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{display_key}_refresh_rate",
                value=str(value) if value is not None else None,
            ),
        )

    async def update_resolution_horizontal(
        self,
        display_key: str,
        value: Optional[int] = None,
    ) -> None:
        """Update resolution horizontal"""
        if value is None:
            value = self._display.sensors_resolution_horizontal(
                self._database, display_key
            )
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{display_key}_resolution_horizontal",
                value=str(value) if value is not None else None,
            ),
        )

    async def update_resolution_vertical(
        self,
        display_key: str,
        value: Optional[int] = None,
    ) -> None:
        """Update resolution vertical"""
        if value is None:
            value = self._display.sensors_resolution_vertical(
                self._database, display_key
            )
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{display_key}_resolution_vertical",
                value=str(value) if value is not None else None,
            ),
        )

    async def update_all_data(self) -> None:
        """Update data"""
        display_list = []
        displays = self._display.sensors_get_displays(self._database)

        if displays is not None and len(displays) > 0:
            # Clear table in case of hardware changes since last run
            self._database.clear_table(DatabaseModel)
            for display_name in displays:
                display_key = make_key(display_name)
                display_list.append(display_key)
                await asyncio.gather(
                    *[
                        self.update_name(display_key, display_name),
                        self.update_pixel_clock(display_key),
                        self.update_refresh_rate(display_key),
                        self.update_resolution_horizontal(display_key),
                        self.update_resolution_vertical(display_key),
                    ]
                )

            if len(display_list) == 0:
                self._logger.debug("No displays found. Using alternative")
                for key, display in enumerate(self._display.get_displays()):
                    display_key = (
                        str(key) if display.name is None else make_key(display.name)
                    )
                    display_list.append(display_key)
                    await asyncio.gather(
                        *[
                            self.update_name(display_key, display.name),
                            self.update_pixel_clock(display_key, display.pixel_clock),
                            self.update_refresh_rate(display_key, display.refresh_rate),
                            self.update_resolution_horizontal(
                                display_key, display.resolution_horizontal
                            ),
                            self.update_resolution_vertical(
                                display_key, display.resolution_vertical
                            ),
                        ]
                    )

            self._database.update_data(
                DatabaseModel,
                DatabaseModel(
                    key="displays",
                    value=dumps(display_list),
                ),
            )



"""System Bridge: Server Handler - Update"""
from typing import Optional

from systembridgeshared.update import Update


def version_update(
    version: str,
) -> dict[str, Optional[str]]:
    """Handle the update request."""
    versions = Update().update(
        version,
        wait=False,
    )
    return versions



"""System Bridge: Update Battery"""
import asyncio

from systembridgeshared.common import camel_to_snake
from systembridgeshared.database import Database
from systembridgeshared.models.database_data import Battery as DatabaseModel

from . import Battery
from ..base import ModuleUpdateBase


class BatteryUpdate(ModuleUpdateBase):
    """Battery Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database)
        self._battery = Battery()

    async def update_sensors(self) -> None:
        """Update Battery Sensors"""
        if data := self._battery.sensors():
            for key, value in data._asdict().items():
                # From status
                if key == "percent":
                    continue
                self._database.update_data(
                    DatabaseModel,
                    DatabaseModel(
                        key="sensors_{key}",
                        value=value,
                    ),
                )

    async def update_status(self) -> None:
        """Update Battery Status"""
        if data := self._battery.status():
            for key, value in data.items():
                self._database.update_data(
                    DatabaseModel,
                    DatabaseModel(
                        key=camel_to_snake(key),
                        value=value,
                    ),
                )

    async def update_all_data(self) -> None:
        """Update data"""
        await asyncio.gather(
            *[
                self.update_sensors(),
                self.update_status(),
            ]
        )



"""System Bridge: Module Base"""

from systembridgeshared.base import Base
from systembridgeshared.database import Database


class ModuleUpdateBase(Base):
    """Module Base"""

    def __init__(
        self,
        database: Database,
    ):
        super().__init__()

        self._database = database

    async def update_all_data(self) -> None:
        """Update data"""
        raise NotImplementedError()



"""System Bridge: Update Memory"""
import asyncio

from systembridgeshared.database import Database
from systembridgeshared.models.database_data import Memory as DatabaseModel

from . import Memory
from ..base import ModuleUpdateBase


class MemoryUpdate(ModuleUpdateBase):
    """Memory Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database)
        self._memory = Memory()

    async def update_swap(self) -> None:
        """Update Swap Memory"""
        for key, value in self._memory.swap()._asdict().items():
            self._database.update_data(
                DatabaseModel,
                DatabaseModel(
                    key=f"swap_{key}",
                    value=value,
                ),
            )

    async def update_virtual(self) -> None:
        """Update Virtual Memory"""
        for key, value in self._memory.virtual()._asdict().items():
            self._database.update_data(
                DatabaseModel,
                DatabaseModel(
                    key=f"virtual_{key}",
                    value=value,
                ),
            )

    async def update_all_data(self) -> None:
        """Update data"""
        await asyncio.gather(
            *[
                self.update_swap(),
                self.update_virtual(),
            ]
        )



"""System Bridge: Server Handler - Power"""
import asyncio
import os
import sys
from collections.abc import Callable


async def schedule_power_event(
    time: int,
    action: Callable[[], None],
) -> None:
    """Schedule a power event."""
    await asyncio.sleep(time)
    action()


def sleep() -> None:
    """Send the system to sleep."""
    if sys.platform == "linux":
        os.system("systemctl suspend")
    elif sys.platform == "win32":
        os.system("rundll32.exe powrprof.dll,SetSuspendState 0,1,0")


def hibernate() -> None:
    """Hibernate the system."""
    if sys.platform == "linux":
        os.system("systemctl hibernate")
    elif sys.platform == "win32":
        os.system("rundll32.exe powrprof.dll,SetSuspendState 1,1,0")


def restart() -> None:
    """Restart the system."""
    if sys.platform == "linux":
        os.system("systemctl restart")
    elif sys.platform == "win32":
        os.system("shutdown /r /t 0")


def shutdown() -> None:
    """Shutdown the system."""
    if sys.platform == "linux":
        os.system("systemctl poweroff")
    elif sys.platform == "win32":
        os.system("shutdown /s /t 0")


def lock() -> None:
    """Lock the system."""
    if sys.platform == "linux":
        os.system("xdg-screensaver lock")
    elif sys.platform == "win32":
        os.system("rundll32.exe user32.dll,LockWorkStation")


def logout() -> None:
    """Logout the user."""
    if sys.platform == "linux":
        os.system("pkexec --user $(whoami) gdm-force-logout")
    elif sys.platform == "win32":
        os.system("logoff")



"""System Bridge: Display"""
from __future__ import annotations

from typing import Optional

from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module
from screeninfo import get_monitors
from systembridgeshared.base import Base
from systembridgeshared.common import make_key
from systembridgeshared.database import Database
from systembridgeshared.models.database_data_sensors import (
    Sensors as SensorsDatabaseModel,
)


class DisplayModel(BaseModel):
    """Display Model"""

    name: str = Field(..., description="Display name")
    pixel_clock: Optional[float] = Field(None, description="Pixel clock")
    refresh_rate: Optional[float] = Field(None, description="Refresh rate")
    resolution_horizontal: int = Field(..., description="Resolution horizontal")
    resolution_vertical: int = Field(..., description="Resolution vertical")


class Display(Base):
    """Display"""

    def get_displays(self) -> list[DisplayModel]:
        """Get Displays"""
        return [
            DisplayModel(
                name=monitor.name if monitor.name is not None else str(key),
                pixel_clock=None,
                refresh_rate=None,
                resolution_horizontal=monitor.width,
                resolution_vertical=monitor.height,
            )
            for key, monitor in enumerate(get_monitors())
        ]

    def sensors_get_displays(
        self,
        database: Database,
    ) -> list[str]:
        """Get Displays"""
        displays = []
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "display" in item.hardware_type.lower()
                and item.hardware_name is not None
                and item.hardware_name not in displays
            ):
                displays.append(item.hardware_name)
        return displays

    def sensors_pixel_clock(
        self,
        database: Database,
        display_key: str,
    ) -> Optional[float]:
        """Display pixel clock"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "display" in item.hardware_type.lower()
                and "pixel" in item.name.lower()
                and "clock" in item.name.lower()
                and make_key(item.hardware_name) == display_key
            ):
                self._logger.debug(
                    "Found display pixel clock: %s = %s",
                    item.key,
                    item.value,
                )
                return item.value
        return None

    def sensors_refresh_rate(
        self,
        database: Database,
        display_key: str,
    ) -> Optional[float]:
        """Display refresh rate"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "display" in item.hardware_type.lower()
                and "refresh" in item.name.lower()
                and "rate" in item.name.lower()
                and make_key(item.hardware_name) == display_key
            ):
                self._logger.debug(
                    "Found display refresh rate: %s = %s",
                    item.key,
                    item.value,
                )
                return item.value
        return None

    def sensors_resolution_horizontal(
        self,
        database: Database,
        display_key: str,
    ) -> Optional[int]:
        """Display resolution horizontal"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "display" in item.hardware_type.lower()
                and "resolution" in item.name.lower()
                and "horizontal" in item.name.lower()
                and make_key(item.hardware_name) == display_key
            ):
                self._logger.debug(
                    "Found display resolution horizontal: %s = %s",
                    item.key,
                    item.value,
                )
                return item.value
        return None

    def sensors_resolution_vertical(
        self,
        database: Database,
        display_key: str,
    ) -> Optional[int]:
        """Display resolution vertical"""
        for item in database.get_data(SensorsDatabaseModel):
            if (
                item.hardware_type is not None
                and "display" in item.hardware_type.lower()
                and "resolution" in item.name.lower()
                and "vertical" in item.name.lower()
                and make_key(item.hardware_name) == display_key
            ):
                self._logger.debug(
                    "Found display resolution vertical: %s = %s",
                    item.key,
                    item.value,
                )
                return item.value
        return None



"""System Bridge: Update GPU"""
import asyncio
from json import dumps

from systembridgeshared.common import make_key
from systembridgeshared.database import Database
from systembridgeshared.models.database_data import GPU as DatabaseModel

from ..base import ModuleUpdateBase
from . import GPU


class GPUUpdate(ModuleUpdateBase):
    """GPU Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database)
        self._gpu = GPU()

    async def update_name(
        self,
        gpu_key: str,
        gpu_name: str,
    ) -> None:
        """Update name"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{gpu_key}_name",
                value=gpu_name,
            ),
        )

    async def update_core_clock(
        self,
        gpu_key: str,
    ) -> None:
        """Update core clock"""
        value = self._gpu.core_clock(self._database, gpu_key)
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{gpu_key}_core_clock",
                value=str(value) if value else None,
            ),
        )

    async def update_core_load(
        self,
        gpu_key: str,
    ) -> None:
        """Update core load"""
        value = self._gpu.core_load(self._database, gpu_key)
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{gpu_key}_core_load",
                value=str(value) if value else None,
            ),
        )

    async def update_fan_speed(
        self,
        gpu_key: str,
    ) -> None:
        """Update fan speed"""
        value = self._gpu.fan_speed(self._database, gpu_key)
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{gpu_key}_fan_speed",
                value=str(value) if value else None,
            ),
        )

    async def update_memory_clock(
        self,
        gpu_key: str,
    ) -> None:
        """Update memory clock"""
        value = self._gpu.memory_clock(self._database, gpu_key)
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{gpu_key}_memory_clock",
                value=str(value) if value else None,
            ),
        )

    async def update_memory_load(
        self,
        gpu_key: str,
    ) -> None:
        """Update memory load"""
        value = self._gpu.memory_load(self._database, gpu_key)
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{gpu_key}_memory_load",
                value=str(value) if value else None,
            ),
        )

    async def update_memory_free(
        self,
        gpu_key: str,
    ) -> None:
        """Update memory free"""
        value = self._gpu.memory_free(self._database, gpu_key)
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{gpu_key}_memory_free",
                value=str(value) if value else None,
            ),
        )

    async def update_memory_used(
        self,
        gpu_key: str,
    ) -> None:
        """Update memory used"""
        value = self._gpu.memory_used(self._database, gpu_key)
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{gpu_key}_memory_used",
                value=str(value) if value else None,
            ),
        )

    async def update_memory_total(
        self,
        gpu_key: str,
    ) -> None:
        """Update memory total"""
        value = self._gpu.memory_total(self._database, gpu_key)
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{gpu_key}_memory_total",
                value=str(value) if value else None,
            ),
        )

    async def update_power(
        self,
        gpu_key: str,
    ) -> None:
        """Update power"""
        value = self._gpu.power(self._database, gpu_key)
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{gpu_key}_power",
                value=str(value) if value else None,
            ),
        )

    async def update_temperature(
        self,
        gpu_key: str,
    ) -> None:
        """Update temperature"""
        value = self._gpu.temperature(self._database, gpu_key)
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key=f"{gpu_key}_temperature",
                value=str(value) if value else None,
            ),
        )

    async def update_all_data(self) -> None:
        """Update data"""
        gpu_list = []
        gpus = self._gpu.get_gpus(self._database)

        if gpus is not None and len(gpus) > 0:
            # Clear table in case of hardware changes since last run
            self._database.clear_table(DatabaseModel)
            for gpu_name in gpus:
                gpu_key = make_key(gpu_name)
                gpu_list.append(gpu_key)
                await asyncio.gather(
                    *[
                        self.update_name(gpu_key, gpu_name),
                        self.update_core_clock(gpu_key),
                        self.update_core_load(gpu_key),
                        self.update_fan_speed(gpu_key),
                        self.update_memory_clock(gpu_key),
                        self.update_memory_load(gpu_key),
                        self.update_memory_free(gpu_key),
                        self.update_memory_used(gpu_key),
                        self.update_memory_total(gpu_key),
                        self.update_power(gpu_key),
                        self.update_temperature(gpu_key),
                    ]
                )
            self._database.update_data(
                DatabaseModel,
                DatabaseModel(
                    key="gpus",
                    value=dumps(gpu_list),
                ),
            )



"""System Bridge: Server Handler - Media"""
from __future__ import annotations

import asyncio
import io
import mimetypes
import os
import re
import tempfile
from collections.abc import Callable
from datetime import datetime
from typing import Optional
from urllib.parse import urlencode

import aiofiles
from aiohttp import ClientSession
from fastapi import status
from fastapi.exceptions import HTTPException
from fastapi.responses import FileResponse
from mutagen._file import File as MutagenFile
from plyer import storagepath
from systembridgeshared.const import (
    QUERY_API_KEY,
    QUERY_API_PORT,
    QUERY_AUTOPLAY,
    QUERY_BASE,
    QUERY_PATH,
    QUERY_URL,
    QUERY_VOLUME,
    SECRET_API_KEY,
    SETTING_ADDITIONAL_MEDIA_DIRECTORIES,
    SETTING_PORT_API,
)
from systembridgeshared.models.media_files import File as MediaFile
from systembridgeshared.models.media_play import MediaPlay
from systembridgeshared.settings import Settings


def get_directories(settings: Settings) -> list[dict[str, str]]:
    """Get directories"""
    directories = [
        {
            "key": "documents",
            "path": storagepath.get_documents_dir(),  # type: ignore
        },
        {
            "key": "downloads",
            "path": storagepath.get_downloads_dir(),  # type: ignore
        },
        {
            "key": "home",
            "path": storagepath.get_home_dir(),  # type: ignore
        },
        {
            "key": "music",
            "path": storagepath.get_music_dir(),  # type: ignore
        },
        {
            "key": "pictures",
            "path": storagepath.get_pictures_dir(),  # type: ignore
        },
        {
            "key": "videos",
            "path": storagepath.get_videos_dir(),  # type: ignore
        },
    ]

    additional_directories = settings.get(SETTING_ADDITIONAL_MEDIA_DIRECTORIES)
    if additional_directories is not None and isinstance(additional_directories, list):
        for directory in additional_directories:
            directories.append(
                {
                    "key": directory["name"],
                    "path": directory["value"],
                }
            )

    return directories


def get_files(
    settings: Settings,
    base_path: str,
    path: str,
) -> list[MediaFile]:
    """Get files from path"""
    root_path = None
    for item in get_directories(settings):
        if item["key"] == base_path:
            root_path = item["path"]
            break

    if root_path is None or not os.path.exists(root_path):
        return []

    files_info = []
    for filename in os.listdir(path):
        file_info = get_file(root_path, os.path.join(path, filename))
        if file_info is not None:
            files_info.append(file_info)

    return files_info


def get_file(
    base_path: str,
    filepath: str,
) -> Optional[MediaFile]:
    """Get file from path"""
    try:
        stat = os.stat(filepath)
    except FileNotFoundError:
        return None

    mime_type = None
    if os.path.isfile(filepath):
        mime_type = mimetypes.guess_type(filepath)[0]

    return MediaFile(
        name=os.path.basename(filepath),
        path=filepath.removeprefix(base_path)[1:],
        fullpath=filepath,
        size=stat.st_size,
        last_accessed=stat.st_atime,
        created=stat.st_ctime,
        modified=stat.st_mtime,
        is_directory=os.path.isdir(filepath),
        is_file=os.path.isfile(filepath),
        is_link=os.path.islink(filepath),
        mime_type=mime_type,
    )


def get_file_data(
    filepath: str,
) -> FileResponse:
    """Get file data"""
    return FileResponse(filepath)


async def write_file(
    filepath: str,
    data: bytes,
) -> None:
    """Write file"""
    async with aiofiles.open(filepath, "wb") as new_file:
        await new_file.write(data)
        await new_file.close()


async def play_media(
    settings: Settings,
    callback: Callable[[str, MediaPlay], None],
    query_autoplay: Optional[bool] = False,
    query_base: Optional[str] = None,
    query_path: Optional[str] = None,
    query_type: Optional[str] = "video",
    query_url: Optional[str] = None,
    query_volume: Optional[float] = 40,
    request_host: Optional[str] = None,
    request_scheme: Optional[str] = "http",
):
    """Handler for media play requests"""
    mime_type = None
    path = None
    if query_url is None:
        if query_base is None:
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST,
                {"message": "No base specified"},
            )
        if query_path is None:
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST,
                {"message": "No path specified"},
            )

        root_path = None
        for item in get_directories(settings):
            if item["key"] == query_base:
                root_path = item["path"]
                break

        if root_path is None or not os.path.exists(root_path):
            raise HTTPException(
                status.HTTP_404_NOT_FOUND,
                {"message": "Cannot find base", "base": query_base},
            )

        if not (path := os.path.join(root_path, query_path)):
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST,
                {"message": "Cannot find path", "path": path},
            )
        if not os.path.exists(path):
            raise HTTPException(
                status.HTTP_404_NOT_FOUND,
                {"message": "File does not exist", "path": path},
            )
        if not os.path.abspath(path).startswith(os.path.abspath(root_path)):
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST,
                {
                    "message": "Path is not underneath base path",
                    "base": root_path,
                    "path": path,
                },
            )
        if not os.path.isfile(path):
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST,
                {"message": "Path is not a file", "path": path},
            )

        query_url = f"""{request_scheme}://{request_host}/api/media/file/data?{urlencode({
                        QUERY_API_KEY: settings.get_secret(SECRET_API_KEY),
                        QUERY_API_PORT: settings.get(SETTING_PORT_API),
                        QUERY_BASE: query_base,
                        QUERY_PATH: query_path,
                    })}"""

        mime_type, _ = mimetypes.guess_type(path)
        if mime_type is None:
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST,
                {"message": "Cannot determine mime type", "path": path},
            )

        query_type = (
            "audio"
            if "audio" in mime_type
            else "video"
            if "video" in mime_type
            else None
        )

        if query_type is None:
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST,
                {
                    "message": "Unsupported file type",
                    "path": path,
                    "mime_type": mime_type,
                },
            )
    else:
        if query_type == "audio":
            path = os.path.join(tempfile.gettempdir(), "tmp.mp3")
            # Download local version to get metadata
            async with ClientSession() as session:
                async with session.get(query_url) as response:
                    data = await response.read()
                    async with aiofiles.open(
                        path,
                        "wb",
                    ) as new_file:
                        await new_file.write(data)
                        await new_file.close()

    media_play = MediaPlay(
        **{
            QUERY_AUTOPLAY: query_autoplay,
            QUERY_URL: query_url,
            QUERY_VOLUME: query_volume,
        }
    )

    api_port = settings.get(SETTING_PORT_API)
    api_key = settings.get_secret(SECRET_API_KEY)

    if query_type == "audio":
        if path is None:
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST,
                {
                    "message": "Failed to get path for audio file",
                },
            )

        if (metadata := MutagenFile(path)) is None:
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST,
                {
                    "message": "Failed to get metadata for audio file",
                },
            )

        album = metadata.get("album")
        if album is not None and len(album) > 0:
            media_play.album = album[0]
        elif (album := metadata.get("TALB")) is not None:
            media_play.album = album.text[0]

        artist = metadata.get("artist")
        if artist is not None and len(artist) > 0:
            media_play.artist = artist[0]
        elif (artist := metadata.get("TPE1")) is not None:
            media_play.artist = artist.text[0]

        title = metadata.get("title")
        if title is not None and len(title) > 0:
            media_play.title = title[0]
        elif (title := metadata.get("TIT2")) is not None:
            media_play.title = title.text[0]

        # MP3 etc.
        for key in metadata.keys():
            if key.startswith("APIC"):
                if (cover := metadata.get(key)) is not None:
                    cover_filename = _save_cover_from_binary(
                        cover.data,
                        cover.mime,
                        media_play.album,
                    )
                    if cover_filename is not None:
                        media_play.cover = f"""{request_scheme}://{request_host}/api/media/file/data?{urlencode({
                                                QUERY_API_KEY: settings.get_secret(SECRET_API_KEY),
                                                QUERY_API_PORT: settings.get(SETTING_PORT_API),
                                                QUERY_BASE: "pictures",
                                                QUERY_PATH: cover_filename,
                                            })}"""
                        asyncio.create_task(_delete_cover_delayed(cover_filename))
                    break

        # FLAC
        if media_play.cover is None:
            try:
                if (
                    getattr(metadata, "pictures") is not None
                    and len(metadata.pictures) > 0
                ):
                    cover_filename = _save_cover_from_binary(
                        metadata.pictures[0].data,
                        metadata.pictures[0].mime,
                        media_play.album,
                    )
                    if cover_filename is not None:
                        media_play.cover = f"""{request_scheme}://{request_host}/api/media/file/data?{urlencode({
                                                QUERY_API_KEY: settings.get_secret(SECRET_API_KEY),
                                                QUERY_API_PORT: settings.get(SETTING_PORT_API),
                                                QUERY_BASE: "pictures",
                                                QUERY_PATH: cover_filename,
                                            })}"""
                        asyncio.create_task(_delete_cover_delayed(cover_filename))
            except AttributeError:
                pass

    if query_type is None:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            {
                "message": "Unsupported file type",
                "path": path,
                "mime_type": mime_type,
            },
        )

    callback(query_type, media_play)

    return {
        "message": "Opened media player",
        "media_type": query_type,
        "mime_type": mime_type,
        "path": path,
        "player_url": f"""{request_scheme}://{request_host}/app/player/{query_type}.html?{urlencode({
                    QUERY_API_KEY: api_key,
                    QUERY_API_PORT: api_port,
                    **media_play.dict(exclude_none=True),
                })}""",
        **media_play.dict(exclude_none=True),
    }


async def _delete_cover_delayed(
    file_name: str,
    delay: float = 20,
) -> None:
    """Delete cover after delay."""
    await asyncio.sleep(delay)
    file_path = os.path.join(
        storagepath.get_pictures_dir(),  # type: ignore
        file_name,
    )
    if os.path.exists(file_path):
        os.remove(file_path)


def _save_cover_from_binary(
    data: bytes,
    mime_type: str,
    name: Optional[str],
) -> str:
    """Save cover from binary."""
    cover_extension = mimetypes.guess_extension(mime_type)
    file_name = re.sub(
        "[^-a-zA-Z0-9_.() ]+",
        "",
        f"{name if name is not None else 'unknown'}__{datetime.timestamp(datetime.now())}{cover_extension}",
    )
    file_path = os.path.join(
        storagepath.get_pictures_dir(),  # type: ignore
        file_name,
    )
    with io.open(file_path, "wb") as cover_file:
        cover_file.write(data)
    return file_name



"""System Bridge: Server Handler - Keyboard"""
from keyboard import press_and_release, write


def keyboard_keypress(key: str):
    """Press a keyboard key"""
    press_and_release(key)


def keyboard_text(text: str):
    """Type text"""
    write(text)



"""System Bridge: Server Handler - Open"""
import os
import subprocess
import sys
from webbrowser import open_new_tab


def open_path(
    path: str,
) -> None:
    """Open a file."""
    if sys.platform == "win32":
        os.startfile(path)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, path])


def open_url(
    url: str,
) -> None:
    """Open a URL in the default browser."""
    open_new_tab(url)



"""System Bridge: Update CPU"""
import asyncio

from systembridgeshared.database import Database
from systembridgeshared.models.database_data import CPU as DatabaseModel

from . import CPU
from ..base import ModuleUpdateBase


class CPUUpdate(ModuleUpdateBase):
    """CPU Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database)
        self._cpu = CPU()

    async def update_count(self) -> None:
        """Update CPU count"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="count",
                value=str(self._cpu.count()),
            ),
        )

    async def update_frequency(self) -> None:
        """Update CPU frequency"""
        for key, value in self._cpu.freq()._asdict().items():
            self._database.update_data(
                DatabaseModel,
                DatabaseModel(
                    key=f"frequency_{key}",
                    value=value,
                ),
            )

    async def update_frequency_per_cpu(self) -> None:
        """Update CPU frequency per CPU"""
        count = 0
        for data in [freq._asdict() for freq in self._cpu.freq_per_cpu()]:
            for key, value in data.items():
                self._database.update_data(
                    DatabaseModel,
                    DatabaseModel(
                        key=f"frequency_{count}_{key}",
                        value=value,
                    ),
                )
            count += 1

    async def update_load_average(self) -> None:
        """Update load average"""
        avg_tuple = self._cpu.load_average()
        result = sum([avg_tuple[0], avg_tuple[1], avg_tuple[2]]) / 3
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="load_average",
                value=str(result),
            ),
        )

    async def update_power_package(self) -> None:
        """Update package power"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="power_package",
                value=str(self._cpu.power_package(self._database)),
            ),
        )

    async def update_power_per_cpu(self) -> None:
        """Update per cpu power"""
        if (result := self._cpu.power_per_cpu(self._database)) is None:
            return None
        for key, value in result:
            self._database.update_data(
                DatabaseModel,
                DatabaseModel(
                    key=f"power_per_cpu_{key}",
                    value=str(value),
                ),
            )

    async def update_stats(self) -> None:
        """Update stats"""
        for key, value in self._cpu.stats()._asdict().items():
            self._database.update_data(
                DatabaseModel,
                DatabaseModel(
                    key=f"stats_{key}",
                    value=value,
                ),
            )

    async def update_temperature(self) -> None:
        """Update temperature"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="temperature",
                value=str(self._cpu.temperature(self._database)),
            ),
        )

    async def update_times(self) -> None:
        """Update times"""
        for key, value in self._cpu.times()._asdict().items():
            self._database.update_data(
                DatabaseModel,
                DatabaseModel(
                    key=f"times_{key}",
                    value=value,
                ),
            )

    async def update_times_percent(self) -> None:
        """Update times percent"""
        for key, value in self._cpu.times_percent()._asdict().items():
            self._database.update_data(
                DatabaseModel,
                DatabaseModel(
                    key=f"times_percent_{key}",
                    value=value,
                ),
            )

    async def update_times_per_cpu(self) -> None:
        """Update times per CPU"""
        count = 0
        for data in [freq._asdict() for freq in self._cpu.times_per_cpu()]:
            for key, value in data.items():
                self._database.update_data(
                    DatabaseModel,
                    DatabaseModel(
                        key=f"times_per_cpu_{count}_{key}",
                        value=value,
                    ),
                )
            count += 1

    async def update_times_per_cpu_percent(self) -> None:
        """Update times per CPU percent"""
        count = 0
        for data in [freq._asdict() for freq in self._cpu.times_per_cpu_percent()]:
            for key, value in data.items():
                self._database.update_data(
                    DatabaseModel,
                    DatabaseModel(
                        key=f"times_per_cpu_percent_{count}_{key}",
                        value=value,
                    ),
                )
            count += 1

    async def update_usage(self) -> None:
        """Update usage"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="usage",
                value=str(self._cpu.usage()),
            ),
        )

    async def update_usage_per_cpu(self) -> None:
        """Update usage per CPU"""
        count = 0
        for value in self._cpu.usage_per_cpu():
            self._database.update_data(
                DatabaseModel,
                DatabaseModel(
                    key=f"usage_{count}",
                    value=str(value),
                ),
            )
            count += 1

    async def update_voltage(self) -> None:
        """Update voltage"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="voltage",
                value=str(self._cpu.voltage(self._database)),
            ),
        )

    async def update_all_data(self) -> None:
        """Update data"""
        await asyncio.gather(
            *[
                self.update_count(),
                self.update_frequency(),
                self.update_frequency_per_cpu(),
                self.update_load_average(),
                self.update_power_package(),
                self.update_power_per_cpu(),
                self.update_stats(),
                self.update_temperature(),
                self.update_times(),
                self.update_times_percent(),
                self.update_times_per_cpu(),
                self.update_times_per_cpu_percent(),
                self.update_usage(),
                self.update_usage_per_cpu(),
                self.update_voltage(),
            ]
        )



"""System Bridge: API"""
import asyncio
import logging
import os
import sys
from collections.abc import Callable
from typing import Any, Optional, Union

from fastapi import Depends, FastAPI, File, Header, Query, Request, WebSocket, status
from fastapi.exceptions import HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse
from fastapi.staticfiles import StaticFiles
from systembridgeshared.common import asyncio_get_loop, convert_string_to_correct_type
from systembridgeshared.const import HEADER_API_KEY, QUERY_API_KEY, SECRET_API_KEY
from systembridgeshared.database import TABLE_MAP, Database
from systembridgeshared.models.data import DataDict
from systembridgeshared.models.database_data_remote_bridge import RemoteBridge
from systembridgeshared.models.keyboard_key import KeyboardKey
from systembridgeshared.models.keyboard_text import KeyboardText
from systembridgeshared.models.media_files import File as MediaFile
from systembridgeshared.models.media_files import MediaFiles
from systembridgeshared.models.media_play import MediaPlay
from systembridgeshared.models.notification import Notification
from systembridgeshared.models.open_path import OpenPath
from systembridgeshared.models.open_url import OpenUrl
from systembridgeshared.settings import Settings

from .._version import __version__
from ..gui import GUI
from ..modules.listeners import Listeners
from ..server.keyboard import keyboard_keypress, keyboard_text
from ..server.media import (
    get_directories,
    get_file,
    get_file_data,
    get_files,
    play_media,
    write_file,
)
from ..server.open import open_path, open_url
from ..server.websocket import WebSocketHandler
from .power import (
    hibernate,
    lock,
    logout,
    restart,
    schedule_power_event,
    shutdown,
    sleep,
)
from .remote_bridge import get_remote_bridges
from .update import version_update

database = Database()
settings = Settings(database)

logger = logging.getLogger("systembridgebackend.server.api")


def security_api_key_header(
    api_key_header: Optional[str] = Header(alias=HEADER_API_KEY, default=None),
):
    """Get API key from request."""
    key = str(settings.get_secret(SECRET_API_KEY))
    if api_key_header is not None and api_key_header == key:
        logger.info("Authorized with API Key Header")
        return True
    return False


def security_api_key_query(
    api_key_query: Optional[str] = Query(alias=QUERY_API_KEY, default=None),
):
    """Get API key from request."""
    key = str(settings.get_secret(SECRET_API_KEY))
    if api_key_query is not None and api_key_query == key:
        logger.info("Authorized with API Key Query Parameter")
        return True
    return False


def security_api_key(
    api_key_header_result: bool = Depends(security_api_key_header),
    api_key_query_result: bool = Depends(security_api_key_query),
):
    """Get API key from request."""
    logger.info("API Key Header Result: %s", api_key_header_result)
    logger.info("API Key Query Result: %s", api_key_query_result)
    if not (api_key_header_result or api_key_query_result):
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Invalid API Key",
        )


class API(FastAPI):
    """Extended FastAPI"""

    def __init__(
        self,
        **kwargs: Any,
    ) -> None:
        """Initialize"""
        super().__init__(**kwargs)
        self.add_middleware(
            CORSMiddleware,
            allow_credentials=True,
            allow_origins="*",
            allow_headers=[
                "accept",
                "api-key",
                "content-type",
                "origin",
            ],
            allow_methods=[
                "DELETE",
                "GET",
                "OPTIONS",
                "POST",
                "PUT",
            ],
        )
        self.callback_exit: Callable[[], None]
        self.callback_open_gui: Callable[[str, str], None]
        self.listeners: Listeners
        self.implemented_modules: list[str] = []
        self.loop: asyncio.AbstractEventLoop = asyncio_get_loop()


app = API(
    title="System Bridge",
    version=__version__.public(),
)


@app.get("/")
def get_root() -> dict[str, str]:
    """Get root."""
    return {
        "message": "Hello!",
    }


@app.get("/api", dependencies=[Depends(security_api_key)])
def get_api_root() -> dict[str, str]:
    """Get API root."""
    return {
        "message": "Hello!",
        "version": __version__.public(),
    }


@app.get("/api/data/{module}", dependencies=[Depends(security_api_key)])
def get_data(module: str) -> DataDict:
    """Get data from module."""
    table_module = TABLE_MAP.get(module)
    if module not in app.implemented_modules or table_module is None:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            detail={"message": f"Data module {module} not found"},
        )
    return database.get_data_dict(table_module)


@app.get("/api/data/{module}/{key}", dependencies=[Depends(security_api_key)])
def get_data_by_key(
    module: str,
    key: str,
) -> dict[str, Any]:
    """Get data from module by key."""
    table_module = TABLE_MAP.get(module)
    if module not in app.implemented_modules or table_module is None:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            detail={"message": f"Data module {module} not found"},
        )
    data = database.get_data_item_by_key(table_module, key)
    if data is None:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            detail={"message": f"Data item {key} in module {module} not found"},
        )
    return {
        data.key: convert_string_to_correct_type(data.value),
        "last_updated": data.timestamp,
    }


@app.post("/api/keyboard", dependencies=[Depends(security_api_key)])
def send_keyboard_event(
    keyboard_event: Union[KeyboardKey, KeyboardText]
) -> dict[str, str]:
    """Send keyboard event."""
    if isinstance(keyboard_event, KeyboardKey):
        try:
            keyboard_keypress(keyboard_event.key)
        except ValueError as error:
            raise HTTPException(
                status.HTTP_400_BAD_REQUEST, detail={"error": str(error)}
            ) from error
        return {
            "message": "Keypress sent",
            **keyboard_event.dict(),
        }
    if isinstance(keyboard_event, KeyboardText):
        keyboard_text(keyboard_event.text)
        return {
            "message": "Text sent",
            **keyboard_event.dict(),
        }
    raise HTTPException(status.HTTP_400_BAD_REQUEST, detail="Invalid keyboard event")


@app.get("/api/media", dependencies=[Depends(security_api_key)])
def get_media_directories() -> dict[str, list[dict[str, str]]]:
    """Get media directories."""
    return {
        "directories": get_directories(settings),
    }


@app.get("/api/media/files", dependencies=[Depends(security_api_key)])
def get_media_files(
    query_base: str = Query(..., alias="base"),
    query_path: Optional[str] = Query(None, alias="path"),
) -> MediaFiles:
    """Get media files."""
    root_path = None
    for item in get_directories(settings):
        if item["key"] == query_base:
            root_path = item["path"]
            break

    if root_path is None or not os.path.exists(root_path):
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            detail={"message": "Cannot find base", "base": query_base},
        )

    path = os.path.join(root_path, query_path) if query_path else root_path
    if not os.path.exists(path):
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            {"message": "Cannot find path", "path": path},
        )
    if not os.path.abspath(path).startswith(os.path.abspath(root_path)):
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            {
                "message": "Path is not underneath base path",
                "base": root_path,
                "path": path,
            },
        )
    if not os.path.isdir(path):
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            {"message": "Path is not a directory", "path": path},
        )

    return MediaFiles(
        files=get_files(settings, query_base, path),
        path=path,
    )


@app.get("/api/media/file", dependencies=[Depends(security_api_key)])
def get_media_file(
    query_base: str = Query(..., alias="base"),
    query_path: str = Query(..., alias="path"),
) -> MediaFile:
    """Get media file info."""
    root_path = None
    for item in get_directories(settings):
        if item["key"] == query_base:
            root_path = item["path"]
            break

    if root_path is None or not os.path.exists(root_path):
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            detail={"message": "Cannot find base", "base": query_base},
        )

    path = os.path.join(root_path, query_path) if query_path else root_path
    if not os.path.exists(path):
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            {"message": "Cannot find path", "path": path},
        )
    if not os.path.abspath(path).startswith(os.path.abspath(root_path)):
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            {
                "message": "Path is not underneath base path",
                "base": root_path,
                "path": path,
            },
        )
    if not os.path.isfile(path):
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            {"message": "Path is not a file", "path": path},
        )

    if (file := get_file(query_base, path)) is None:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            {"message": "Cannot get file", "path": path},
        )
    return file


@app.get("/api/media/file/data", dependencies=[Depends(security_api_key)])
def get_media_file_data(
    query_base: str = Query(..., alias="base"),
    query_path: str = Query(..., alias="path"),
) -> FileResponse:
    """Get media file data."""
    root_path = None
    for item in get_directories(settings):
        if item["key"] == query_base:
            root_path = item["path"]
            break

    if root_path is None or not os.path.exists(root_path):
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            detail={"message": "Cannot find base", "base": query_base},
        )

    path = os.path.join(root_path, query_path) if query_path else root_path
    if not os.path.exists(path):
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            {"message": "Cannot find path", "path": path},
        )
    if not os.path.abspath(path).startswith(os.path.abspath(root_path)):
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            {
                "message": "Path is not underneath base path",
                "base": root_path,
                "path": path,
            },
        )
    if not os.path.isfile(path):
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            {"message": "Path is not a file", "path": path},
        )

    return get_file_data(path)


@app.post("/api/media/file/write", dependencies=[Depends(security_api_key)])
async def send_media_file(
    query_base: str = Query(..., alias="base"),
    query_path: str = Query(..., alias="path"),
    query_filename: str = Query(..., alias="filename"),
    file: bytes = File(...),
) -> dict[str, str]:
    """Send media file."""
    root_path = None
    for item in get_directories(settings):
        if item["key"] == query_base:
            root_path = item["path"]
            break

    if root_path is None or not os.path.exists(root_path):
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            detail={"message": "Cannot find base", "base": query_base},
        )

    path = os.path.join(root_path, query_path) if query_path else root_path
    if not os.path.exists(path):
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            {"message": "Cannot find path", "path": path},
        )
    if not os.path.abspath(path).startswith(os.path.abspath(root_path)):
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            {
                "message": "Path is not underneath base path",
                "base": root_path,
                "path": path,
            },
        )

    await write_file(os.path.join(path, query_filename), file)

    return {
        "message": "File uploaded",
        "path": path,
        "filename": query_filename,
    }


@app.post("/api/media/play", dependencies=[Depends(security_api_key)])
async def send_media_play(
    request: Request,
    query_autoplay: Optional[bool] = Query(False, alias="autoplay"),
    query_base: Optional[str] = Query(None, alias="base"),
    query_path: Optional[str] = Query(None, alias="path"),
    query_type: Optional[str] = Query(None, alias="type"),
    query_url: Optional[str] = Query(None, alias="url"),
    query_volume: Optional[float] = Query(40, alias="volume"),
) -> dict[str, str]:
    """Play media."""
    return await play_media(
        settings,
        callback_media_play,
        query_autoplay=query_autoplay,
        query_base=query_base,
        query_path=query_path,
        query_type=query_type,
        query_url=query_url,
        query_volume=query_volume,
        request_host=request.url.hostname,
        request_scheme=request.url.scheme,
    )


@app.post("/api/notification", dependencies=[Depends(security_api_key)])
def send_notification(notification: Notification) -> dict[str, str]:
    """Send notification."""
    app.callback_open_gui("notification", notification.json())
    return {"message": "Notification sent"}


@app.post("/api/open", dependencies=[Depends(security_api_key)])
def send_open(open_model: Union[OpenPath, OpenUrl]) -> dict[str, str]:
    """Send notification."""
    if isinstance(open_model, OpenPath) and open_model.path is not None:
        open_path(open_model.path)
        return {
            "message": f"Opening path: {open_model.path}",
        }
    if isinstance(open_model, OpenUrl) and open_model.url is not None:
        open_url(open_model.url)
        return {
            "message": f"Opening URL: {open_model.url}",
        }
    raise HTTPException(
        status.HTTP_400_BAD_REQUEST,
        {"message": "No path or URL specified"},
    )


@app.post("/api/power/sleep", dependencies=[Depends(security_api_key)])
def send_power_sleep() -> dict[str, str]:
    """Send power sleep."""
    app.loop.create_task(
        schedule_power_event(2, sleep),
        name="Power Sleep",
    )
    return {"message": "Sleeping"}


@app.post("/api/power/hibernate", dependencies=[Depends(security_api_key)])
def send_power_hibernate() -> dict[str, str]:
    """Send power hibernate."""
    app.loop.create_task(
        schedule_power_event(2, hibernate),
        name="Power Hibernate",
    )
    return {"message": "Hibernating"}


@app.post("/api/power/restart", dependencies=[Depends(security_api_key)])
def send_power_restart() -> dict[str, str]:
    """Send power restart."""
    app.loop.create_task(
        schedule_power_event(2, restart),
        name="Power Restart",
    )
    return {"message": "Restarting"}


@app.post("/api/power/shutdown", dependencies=[Depends(security_api_key)])
def send_power_shutdown() -> dict[str, str]:
    """Send power shutdown."""
    app.loop.create_task(
        schedule_power_event(2, shutdown),
        name="Power Shutdown",
    )
    return {"message": "Shutting down"}


@app.post("/api/power/lock", dependencies=[Depends(security_api_key)])
def send_power_lock() -> dict[str, str]:
    """Send power lock."""
    app.loop.create_task(
        schedule_power_event(2, lock),
        name="Power Lock",
    )
    return {"message": "Locking"}


@app.post("/api/power/logout", dependencies=[Depends(security_api_key)])
def send_power_logout() -> dict[str, str]:
    """Send power logout."""
    app.loop.create_task(
        schedule_power_event(2, logout),
        name="Power Logout",
    )
    return {"message": "Logging out"}


@app.delete("/api/remote/{key}", dependencies=[Depends(security_api_key)])
def delete_remote(key: str) -> dict[str, Union[dict, str]]:
    """Delete remote bridge."""
    bridges: list[RemoteBridge] = get_remote_bridges(database)
    remote_bridge: Optional[RemoteBridge] = None

    for bridge in bridges:
        if bridge.key == key:
            remote_bridge = bridge

    if remote_bridge is None:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            {
                "message": "Remote bridge not found",
            },
        )

    database.delete_remote_bridge(remote_bridge.key)

    return {
        "message": "Deleted remote bridge",
        "data": remote_bridge.dict(),
    }


@app.get("/api/remote", dependencies=[Depends(security_api_key)])
def get_remote() -> dict[str, Union[list[dict], str]]:
    """Get remote bridges."""
    return {
        "message": "Got remote bridges",
        "data": [bridge.dict() for bridge in get_remote_bridges(database)],
    }


@app.post("/api/remote", dependencies=[Depends(security_api_key)])
def send_remote(remote: RemoteBridge) -> dict[str, Union[dict, str]]:
    """Send remote bridge."""
    database.update_remote_bridge(remote)
    return {
        "message": "Added remote bridge",
        "data": remote.dict(),
    }


@app.put("/api/remote/{key}", dependencies=[Depends(security_api_key)])
def update_remote(key: str, remote: RemoteBridge) -> dict[str, Union[dict, str]]:
    """Update remote bridge."""
    bridges: list[RemoteBridge] = get_remote_bridges(database)
    remote_bridge: Optional[RemoteBridge] = None

    for bridge in bridges:
        if bridge.key == key:
            remote_bridge = bridge

    if remote_bridge is None:
        raise HTTPException(
            status.HTTP_404_NOT_FOUND,
            {
                "message": "Remote bridge not found",
            },
        )

    database.update_remote_bridge(remote)

    return {
        "message": "Updated remote bridge",
        "data": remote.dict(),
    }


@app.post("/api/update", dependencies=[Depends(security_api_key)])
def send_update(
    query_version: str = Query(..., alias="version")
) -> dict[str, Union[dict[str, Optional[str]], str]]:
    """Send update."""
    if (versions := version_update(query_version)) is None:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            {"message": "Invalid version"},
        )
    return {
        "message": "Updating the application",
        "versions": versions,
    }


@app.websocket("/api/websocket")
async def websocket_endpoint(websocket: WebSocket):
    """Websocket endpoint."""
    await websocket.accept()
    websocket_handler = WebSocketHandler(
        database,
        settings,
        app.listeners,
        app.implemented_modules,
        websocket,
        app.callback_exit,
        app.callback_open_gui,
    )
    await websocket_handler.handler()


if "--no-frontend" not in sys.argv:
    try:
        # pylint: disable=import-error, import-outside-toplevel
        from systembridgefrontend import get_frontend_path

        frontend_path = get_frontend_path()
        logger.info("Serving frontend from: %s", frontend_path)
        app.mount(
            path="/",
            app=StaticFiles(directory=frontend_path),
            name="Frontend",
        )
    except (ImportError, ModuleNotFoundError) as err:
        logger.error("Frontend not found: %s", err)


def callback_media_play(
    media_type: str,
    media_play: MediaPlay,
) -> None:
    """Callback to open media player"""
    gui_player = GUI(settings)
    app.loop.create_task(
        gui_player.start(
            app.callback_exit,
            "media-player",
            media_type,
            media_play.json(),
        ),
        name="GUI media player",
    )



"""System Bridge: Update System"""
import asyncio

from systembridgeshared.database import Database
from systembridgeshared.models.database_data import System as DatabaseModel

from . import System
from ..base import ModuleUpdateBase


class SystemUpdate(ModuleUpdateBase):
    """System Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database)
        self._system = System()

    async def update_boot_time(self) -> None:
        """Update boot time"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="boot_time",
                value=str(self._system.boot_time()),
            ),
        )

    async def update_fqdn(self) -> None:
        """Update FQDN"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="fqdn",
                value=self._system.fqdn(),
            ),
        )

    async def update_hostname(self) -> None:
        """Update hostname"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="hostname",
                value=self._system.hostname(),
            ),
        )

    async def update_ip_address_4(self) -> None:
        """Update IP address 4"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="ip_address_4",
                value=self._system.ip_address_4(),
            ),
        )

    async def update_mac_address(self) -> None:
        """Update MAC address"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="mac_address",
                value=self._system.mac_address(),
            ),
        )

    async def update_platform(self) -> None:
        """Update platform"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="platform",
                value=self._system.platform(),
            ),
        )

    async def update_platform_version(self) -> None:
        """Update platform version"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="platform_version",
                value=self._system.platform_version(),
            ),
        )

    async def update_uptime(self) -> None:
        """Update uptime"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="uptime",
                value=str(self._system.uptime()),
            ),
        )

    async def update_users(self) -> None:
        """Update users"""
        for user in self._system.users():
            for key, value in user._asdict().items():
                self._database.update_data(
                    DatabaseModel,
                    DatabaseModel(
                        key=f"user_{user.name.replace(' ','_').lower()}_{key}",
                        value=value,
                    ),
                )

    async def update_uuid(self) -> None:
        """Update UUID"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="uuid",
                value=self._system.uuid(),
            ),
        )

    async def update_version(self) -> None:
        """Update version"""
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="version",
                value=self._system.version(),
            ),
        )

    async def update_version_latest(self) -> None:
        """Update latest version"""
        release = await self._system.version_latest()
        if release and release.tag_name:
            self._database.update_data(
                DatabaseModel,
                DatabaseModel(
                    key="version_latest",
                    value=release.tag_name.replace("v", "")
                    if release is not None
                    else None,
                ),
            )

    async def update_version_newer_available(self) -> None:
        """Update newer version available"""
        value = self._system.version_newer_available(self._database)
        self._database.update_data(
            DatabaseModel,
            DatabaseModel(
                key="version_newer_available",
                value=str(value) if value else str(False),
            ),
        )

    async def update_all_data(self) -> None:
        """Update data"""
        await asyncio.gather(
            *[
                self.update_boot_time(),
                self.update_fqdn(),
                self.update_hostname(),
                self.update_ip_address_4(),
                self.update_mac_address(),
                self.update_platform(),
                self.update_platform_version(),
                self.update_uptime(),
                self.update_users(),
                self.update_uuid(),
                self.update_version(),
                self.update_version_latest(),
            ]
        )
        # Run after other version updates
        await self.update_version_newer_available()



"""System Bridge: Server"""
import asyncio
import sys
from collections.abc import Callable
from typing import Optional

import uvicorn
from systembridgeshared.base import Base
from systembridgeshared.const import SETTING_LOG_LEVEL, SETTING_PORT_API
from systembridgeshared.database import Database
from systembridgeshared.settings import Settings

from ..data import Data
from ..gui import GUI
from ..modules.listeners import Listeners
from ..server.mdns import MDNSAdvertisement
from .api import app as api_app


class APIServer(uvicorn.Server):
    """Customized uvicorn.Server

    Uvicorn server overrides signals and we need to include
    Tasks to the signals."""

    def __init__(
        self,
        config: uvicorn.Config,
        exit_callback: Callable[[], None],
    ) -> None:
        super().__init__(config)
        self._exit_callback = exit_callback

    def handle_exit(self, sig: int, frame) -> None:
        """Handle exit."""
        self._exit_callback()
        return super().handle_exit(sig, frame)


class Server(Base):
    """Server"""

    def __init__(
        self,
        database: Database,
        settings: Settings,
        listeners: Listeners,
        implemented_modules: list[str],
    ) -> None:
        """Initialize"""
        super().__init__()
        self._gui_notification: Optional[GUI] = None
        self._gui_player: Optional[GUI] = None
        self._gui: Optional[GUI] = None
        self._listeners = listeners
        self._settings = settings
        self._tasks: list[asyncio.Task] = []

        self._mdns_advertisement = MDNSAdvertisement(settings)
        self._mdns_advertisement.advertise_server()

        self._logger.info("Setup API app")
        api_app.callback_exit = self.exit_application
        api_app.callback_open_gui = self.callback_open_gui
        api_app.listeners = listeners
        api_app.implemented_modules = implemented_modules
        api_app.loop = asyncio.get_event_loop()

        self._logger.info("Setup API server")
        self._api_server = APIServer(
            config=uvicorn.Config(
                api_app,
                host="0.0.0.0",
                loop="asyncio",
                log_config=None,
                log_level=str(settings.get(SETTING_LOG_LEVEL)).lower(),
                port=int(str(settings.get(SETTING_PORT_API))),
                workers=4,
            ),
            exit_callback=self.exit_application,
        )
        self._data = Data(database, self.callback_data_updated)
        self._logger.info("Server initialized")

    async def start(self) -> None:
        """Start the server"""
        self._logger.info("Start server")
        self._tasks.extend(
            [
                asyncio.create_task(
                    self._api_server.serve(),
                    name="API",
                ),
                asyncio.create_task(
                    self.update_data(),
                    name="Update data",
                ),
                asyncio.create_task(
                    self.update_frequent_data(),
                    name="Update frequent data",
                ),
            ]
        )
        if "--no-gui" not in sys.argv:
            self._gui = GUI(self._settings)
            self._tasks.append(
                asyncio.create_task(
                    self._gui.start(self.exit_application),
                    name="GUI",
                )
            )

        await asyncio.wait(self._tasks)

    async def callback_data_updated(
        self,
        module: str,
    ) -> None:
        """Data updated"""
        await self._listeners.refresh_data_by_module(module)

    def callback_open_gui(
        self,
        command: str,
        data: str,
    ) -> None:
        """Open GUI"""
        if command == "notification":
            if self._gui_notification:
                self._gui_notification.stop()
            self._gui_notification = GUI(self._settings)
            self._tasks.append(
                asyncio.create_task(
                    self._gui_notification.start(
                        self.exit_application,
                        command,
                        data,
                    ),
                    name="GUI Notification",
                )
            )
        elif command == "player":
            if self._gui_player:
                self._gui_player.stop()
            self._gui_player = GUI(self._settings)
            self._tasks.append(
                asyncio.create_task(
                    self._gui_player.start(
                        self.exit_application,
                        command,
                        data,
                    ),
                    name="GUI Media Player",
                )
            )

    def exit_application(self) -> None:
        """Exit application"""
        self._logger.info("Exiting application")
        for task in self._tasks:
            task.cancel()
        self._logger.info("Tasks cancelled")
        if self._gui:
            self._gui.stop()
        if self._gui_notification:
            self._gui_notification.stop()
        if self._gui_player:
            self._gui_player.stop()
        self._logger.info("GUI stopped. Exiting Application")
        sys.exit(0)

    async def update_data(self) -> None:
        """Update data"""
        self._logger.info("Update data")
        self._data.request_update_data()
        self._logger.info("Schedule next update in 2 minutes")
        await asyncio.sleep(120)
        await self.update_data()

    async def update_frequent_data(self) -> None:
        """Update frequent data"""
        self._logger.info("Update frequent data")
        self._data.request_update_frequent_data()
        self._logger.info("Schedule next frequent update in 30 seconds")
        await asyncio.sleep(30)
        await self.update_frequent_data()



"""System Bridge: Main"""
import asyncio
import logging
import os
import sys

from systembridgeshared.base import Base
from systembridgeshared.const import SETTING_AUTOSTART, SETTING_LOG_LEVEL
from systembridgeshared.database import Database
from systembridgeshared.logger import setup_logger
from systembridgeshared.settings import Settings

from .autostart import autostart_disable, autostart_enable
from .modules.listeners import Listeners
from .modules.system import System
from .server import Server
from .shortcut import create_shortcuts


class Main(Base):
    """Main"""

    def __init__(self) -> None:
        """Initialize"""
        super().__init__()
        if "--init" in sys.argv:
            self._logger.info("Initialized application. Exiting now.")
            sys.exit(0)

        self._logger.info("System Bridge %s: Startup", System().version())

        if "--cli" not in sys.argv:
            autostart = settings.get(SETTING_AUTOSTART)
            self._logger.info("Autostart enabled: %s", autostart)
            if autostart:
                autostart_enable()
            else:
                autostart_disable()

            create_shortcuts()

        implemented_modules = []
        for _, dirs, _ in os.walk(os.path.join(os.path.dirname(__file__), "./modules")):
            implemented_modules = list(filter(lambda d: "__" not in d, dirs))
            break

        listeners = Listeners(database, implemented_modules)

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        self._server = Server(
            database,
            settings,
            listeners,
            implemented_modules,
        )
        loop.run_until_complete(self._server.start())


if __name__ == "__main__":

    database = Database()
    settings = Settings(database)

    LOG_LEVEL = str(settings.get(SETTING_LOG_LEVEL))
    logger = setup_logger(LOG_LEVEL, "system-bridge")
    logging.getLogger("zeroconf").setLevel(logging.ERROR)

    try:
        Main()
    except Exception as exception:  # pylint: disable=broad-except
        logger.fatal("Unhandled error in application", exc_info=exception)
